#include "cygwin.h"
#include <string.h>
#include "donkey.h"
#include "misc.h"
#include "BTree.h"
#include <db.h>
void        Cpy_dtype (      void *A, const void *B) {{{
	ASSERT (A != NULL);
	ASSERT (B != NULL);
	memcpy(A, B, sizeof (dtype));
	}}}
const char *Str_dtype (const void *A) {{{
	const dtype *a = (const dtype*)A;
static char buf[128];
	if (a == NULL) {
		sprintf(buf, "(null)");
		return buf;
		}
	sprintf(buf, "[%9u - %9u](%1u) {%9lu}", a->von, a->bis, a->stat, a->disk);
	return buf;
	}}}

bool auto_sync = true;

DB          *  initBtree (const char *name) {{{
	DB *ret = NULL;
	int err = 0;
	DB_create(ret);
	DB_set_bt_compare2 (ret ,cmp_u32);
//	DB_set_cachesize2 (ret, db_cache_size);
	ret->set_errfile(ret, stdout);
	DB_set_alloc(ret);
	DB_open2 (ret, name, "part", DB_BTREE, DB_CREATE, 0600);
	return ret;
	}}}
void         updateBtree (DB *db, const dtype *d) {{{
	ASSERT (d->von <= d->bis);
	dtype x;
	memcpy(&x, d, sizeof(dtype));
	int err = 0;
	INIT_DBT(key, &x.von, sizeof(x.von));
	INIT_DBT(val, &x    , sizeof(x    ));
	DB_put2 (db, key, val);
	}}}
void         deleteBtree (DB *db, const dtype *d) {{{
	int err = 0;
	uint32_t von = d->von;
	INIT_DBT(key, &von, sizeof(von));
	if (0 != (err = db->del (db, NULL, &key, 0))) {
		if (err == DB_NOTFOUND) return;
		db_error (__FILE__, __LINE__, err);
		}
	}}}
const dtype *searchBtree (DB *db, uint32_t von) {{{
static  dtype ret;
	bzero (&ret, sizeof(ret));
	uint32_t key_von = von;
	INIT_DBT(key, &key_von, sizeof(key_von));
	INIT_DBT(val, &ret    , sizeof(ret    ));
	int err = 0;
	if (0 != (err = DB_get (db, key, val))) {
	if (DB_NOTFOUND != err) db_error (__FILE__, __LINE__, err);
		return NULL;
		}
	return (dtype*) val.data;
	}}}
void          closeBtree (DB *db) {{{
	int err = 0;
	db->sync (db ,0);
	DB_close2(db);
	}}}
const dtype *  Btree_greater_equal (DB *db, uint32_t von, int code) {{{
	// code:
	// 0: set_range
	// 1: next
	// 2: prev
	// 3: ende
	int err;
static	DBC *akt = NULL;
static	DBT key;	// DB_DBT_USERMEM
static	DBT val;	// DB_DBT_USERMEM
static  dtype ret;
static	uint32_t key_val;
	bzero (&ret, sizeof(ret));
	int direction = DB_NEXT;

	switch(code) {
	case 0: ASSERT(akt == NULL);
		key_val = von;
		RESET_DBT (key, &key_val, sizeof(key_val));
		RESET_DBT (val, &ret    , sizeof(ret    ));
		DB_cursor2(db, akt);
		if (0 != (err = akt->c_get (akt, &key, &val, DB_SET_RANGE))) {
			if (err != DB_NOTFOUND) db_error (__FILE__, __LINE__, err);
			if (0 != (err = akt->c_get (akt, &key, &val, DB_LAST))) db_error (__FILE__, __LINE__, err);
			}
		if (ret.von > von) 
			if (0 != (err = akt->c_get (akt, &key, &val, DB_PREV))) db_error (__FILE__, __LINE__, err);
		ASSERT (ret.von <= von && ret.bis >= von);
		return &ret;
	case 1:	direction = DB_NEXT; break;
	case 2:	direction = DB_PREV; break;
	case 4:	direction = DB_LAST; break;
	case 3: if (akt == NULL) return NULL;
		DB_c_close(akt);
		akt = NULL;
		return NULL;
	default:ASSERT(0);		break;
		}
	ASSERT(akt != NULL);
	if (0 != (err = akt->c_get (akt, &key, &val, direction))) {
		if (err != DB_NOTFOUND) db_error (__FILE__, __LINE__, err);
		return NULL;
		}
	return &ret;
	}}}
const dtype *  nextBtree (DB *db, bool first) {{{
	int err;
static  DBC *akt = NULL;
static  DBT key;	// DB_DBT_USERMEM
static  DBT val;	// DB_DBT_USERMEM
static  dtype ret;
static  uint32_t key_val;
	if (akt != NULL && first) {
		DB_c_close(akt);
		akt = NULL;
		}
	if (akt == NULL && !first) {
		ASSERT(0);
		return NULL;
		}
	if (akt == NULL) {
		DB_cursor2 (db, akt);
		bzero (&ret, sizeof(ret));
		key_val = 0;
		RESET_DBT(key, &key_val, sizeof(key_val));		
		RESET_DBT(val, &ret    , sizeof(ret    ));		
		if (0 != (err = akt->c_get (akt, &key, &val, DB_FIRST))) {
			if (err != DB_NOTFOUND) db_error (__FILE__, __LINE__, err);
			DB_c_close(akt);
			akt = NULL;
			return NULL;
			}
//		printf("FIRST FOUND\n");
		return &ret;
		}
	if (0 != (err = akt->c_get (akt, &key, &val, DB_NEXT))) {
		if (err != DB_NOTFOUND) db_error (__FILE__, __LINE__, err);
		DB_c_close(akt);
		akt = NULL;
		return NULL;
		}
//	printf("NEXT FOUND\n");
	return &ret;
	}}}
unsigned countBtree  (DB *db) {{{
	int err;
	DB_BTREE_STAT *statistik;
	DB_stat(db, statistik);
	uint32_t ret = statistik->bt_nkeys;
	free (statistik);
	return ret;
	}}}
void     Print_range (DB *db) {{{
	dtype r;
	r.since = currentTime;
	r.stat = BLOCK_LOCALE;
	r.disk = 0;
	const dtype *akt = nextBtree (db , true);
	while (akt != NULL) {
//		printf("-> [%9u-%9u] %4u <%9u>\r\n", akt->von, akt->bis, akt->stat,akt->bis - akt->von + 1);
		akt = nextBtree (db , false);
		}
	}}}
void    loadBtree    (DB *db, int ,off_t ,uint32_t size) {{{	// DB_DBT_USERMEM
	uint32_t count = countBtree (db);
	if (count == 0) {{{
		dtype r;
		r.since = currentTime;
		r.stat = 0;
		r.disk = 0;
		r.von = 0;
		r.bis = size;
		updateBtree (db, &r);
		db->sync(db, 0);
		return;
		}}}
	int err;
	DBC *akt = NULL;
	DB_cursor2(db, akt);
	uint32_t von_key = 0;
	dtype ret;
	bzero (&ret, sizeof(ret));
	INIT_DBT(key, &von_key, sizeof(von_key));
	INIT_DBT(val, &ret    , sizeof(ret    ));
	uint32_t von_1 = 0;
	while(0 == akt->c_get (akt, &key, &val, DB_NEXT)) {
		if (von_1 < ret.von) {
			dtype oops;
			oops.von   = von_1;
			oops.bis   = ret.bis - 1;
			oops.stat  = BLOCK_LOCALE;
			oops.since = 0;
			oops.disk  = 0;
			INIT_DBT(o_k, &oops.von, sizeof(oops.von));
			INIT_DBT(o_v, &oops    , sizeof(oops    ));
			DB_put2 (db, o_k, o_v);
			}
		von_1 = ret.bis + 1;
		if (ret.stat == BLOCK_LOCALE) continue;
		ret.stat  = 0;
		ret.since = 0;
		DB_put2 (db, key, val);
		}
	DB_c_close(akt);
	db->sync(db, 0);
	fflush(stdout);
	}}}
int     Change_range (DB *db, dtype &n) {{{ // DB_DBT_USERMEM
	if (db == NULL) {
		printf("ERROR db == NULL\r\n");
		return 0;
		}
	if (n.von > n.bis) {
		if (n.bis < 10) n.von = 0;
		else		n.von = n.bis - 10;
		}
	ASSERT (n.von <= n.bis);
	dtype ret;
	bzero (&ret, sizeof(ret));
	DBC *akt = NULL;
	unsigned orig_von = n.von;
	int err = 0;
	DB_cursor2(db, akt);
	INIT_DBT(val    , &ret     , sizeof(ret));
	INIT_DBT(akt_key, &orig_von, sizeof(orig_von));

	if (0 != (err = akt->c_get (akt, &akt_key, &val, DB_SET_RANGE))) {{{ // if here is no block, than the last one is <= von
		if (err != DB_NOTFOUND) db_error (__FILE__, __LINE__, err);
		if (0 != (err = akt->c_get (akt, &akt_key, &val, DB_LAST))) {{{ // first Block in the DB ?
			if (err != DB_NOTFOUND) db_error (__FILE__, __LINE__, err);
			INIT_DBT(val_key, &n.von, sizeof(n.von));
			INIT_DBT(val_dat, &n    , sizeof(n    ));
			DB_put2 (db, val_key, val_dat);
			DB_c_close(akt);
			return 0;
			}}}
		}}}
	else {{{	// if we are to big than only one else there is an block X.von > n.von && X.von < ret.von INVALID
		if (ret.von > n.von)
			if (0 != (err = akt->c_get (akt, &akt_key, &val, DB_PREV))) { // we are smaler than all others
				if (err != DB_NOTFOUND) db_error (__FILE__, __LINE__, err);
				INIT_DBT(val_key, &n.von, sizeof(n.von));
				INIT_DBT(val_dat, &n    , sizeof(n    ));
				DB_put2 (db, val_key, val_dat);
				DB_c_close(akt);
				return 0;
				}
		}}}
	if (ret.bis < n.von) {{{ // Bigger than all other
		INIT_DBT(val_key, &n.von, sizeof(n.von));
		INIT_DBT(val_dat, &n    , sizeof(n    ));
		DB_put2 (db, val_key, val_dat);
		DB_c_close(akt);
		return 0;
		}}}
	ASSERT (ret.von <= n.von);
	bool start = false;
	bool ende  = false;
	if (n.von >= ret.von && n.bis <= ret.bis && n.stat == ret.stat) {
		DB_c_close(akt);
		return 0;
		}
	do {
		INIT_DBT(val_key, &ret.von, sizeof(ret.von));
		DB_del2 (db, val_key);
		if (ret.von < n.von && ret.stat != n.stat) { // prefix with other type => begin found
			unsigned tmp = ret.bis;
			ret.bis = n.von - 1;
			DB_put2 (db, val_key, val);
			ret.bis = tmp;
			start = true;
			}
		if (ret.bis > n.bis && ret.stat != n.stat) { // prefix with other type => end found
			int tmp = ret.von;
			ret.von = n.bis + 1;
			DB_put2 (db, val_key, val);
			ret.von = tmp;
			ende = true;
			}
		if (ret.von < n.von && ret.stat == n.stat) n.von = ret.von;	
		if (ret.bis > n.bis && ret.stat == n.stat) n.bis = ret.bis;	
		if (!start) {{{ // maybe block bevore can be joined
			if (0 != (err = akt->c_get (akt, &akt_key, &val, DB_PREV))) {
				if (err != DB_NOTFOUND) db_error (__FILE__, __LINE__, err);
				start = true;
				}
			else continue;
			}}}
		if (!ende ) {{{ // maybe block after can be joined
			if (0 != (err = akt->c_get (akt, &akt_key, &val, DB_NEXT))) {
				if (err != DB_NOTFOUND) db_error (__FILE__, __LINE__, err);
				ende = true;
				}
			else continue;
			}}}
		} while(! (start && ende));
	INIT_DBT(val_key, &n.von, sizeof(n.von));
	val.size = sizeof(n);
	val.data = &n;
	DB_put2 (db, val_key, val);
	if (n.stat == 0 || n.stat == 2000) db->sync(db, 0);
	if (n.stat == 2000) {
		unsigned block = orig_von / 9728000;
		unsigned VON = block * 9728000;
		unsigned BIS = VON   + 9728000 - 1;
		if (n.von <= VON && n.bis >= BIS) return 1;
		}
	DB_c_close(akt);
	return 0;
	}}}

// vim600: fdm=marker
