#ifndef cSocket_h
#define cSocket_h
#include <time.h>
#include "cygwin.h"	// here only for "struct in_addr"
#include "donkey.h"

#ifndef _WIN32
#define INVALID_SOCKET  (-1)
#define SOCKET		int
#endif

extern size_t	 dlSecond;
extern size_t	 ulSecond;
extern size_t	 global_ul_count;
extern size_t	 global_dl_count;
extern int	 ul_left;
extern int	 dl_left;
extern struct in_addr	 my_ip;
extern char	*socket_db;

typedef enum {
	tNone   	=   0,
	tSocket		=   4,
	tSource 	=   8,
	tServer 	=  16,
	tGui		=  32,
	tCommand	=  64,
	tKademlia 	= 128
	} listType;

class cSocket;


template <class T> class cDLL {
private:
	cDLL<T>	*prev;
	cDLL<T>	*next;
	T		*data;
private:		 cDLL (cDLL<T> *after, T *neu);
public: 		 cDLL () { prev = next = this; data = NULL; }
			~cDLL ();
	cDLL<T>	*add_front	(T *neu);
	cDLL<T>	*add_back	(T *neu);
	bool		 empty		(void) { return next == this; }
	T 		*peek		(void) { return next->data; }
static	void		 remove		(cDLL<T> **old);
	};

typedef cDLL<cSocket> cSocketList;

extern cSocketList timer_list;

class cSocket {
	public: // test only
	SOCKET		 fd;		// Socket handle (My_IP is missing)
	cSocketList		*event;
	cSocketList 	*timer;
	time_t		   work_last;
	private:
	SOCKET		 fd_new;	// Socket on new connection ( cSocket(cSocket *) NOT cSocket(int *fd) )
	public:
//	class cSocket	*nextEvent;
	listType	 Type;
	size_t    	 dl_count;
	size_t    	 ul_count;
	unsigned	 mask;		// Mask wat to do

inline	bool		 validSocket (void) const { return (fd != INVALID_SOCKET); }
	int		 prio;		// 0=Senden ohne account als erstes
					// 1=Normal 
					// rest kann definiert werden
	int		 last_err;
	int		 type;		// UDP/TCP
	int		 connected;	// TCP established
	int		 ready_send;	// send buffer is not full
	struct in_addr 	 this_ip;	// The local  ip
	unsigned short	 this_port;	// The local  port
	struct in_addr	 peer_ip;	// the client ip
	unsigned short	 peer_port;	// The client port
	time_t		 since;		// Time the socket was created
	time_t		 last_conn;	// Time since socket is connected
	time_t		 last_send;	// Time i recived last message
	time_t		 last_read;	// Time i send last message
	char		*send_buf;
	size_t		 send_len;
	char		*read_buf;
	size_t		 read_len;
			 cSocket (class cSocket *new_fd);
			 cSocket (unsigned short port, int type);
			 cSocket (unsigned int   ip  , unsigned short port, int type);
virtual			~cSocket ();
	bool		 Write 	 (const  char *buf, size_t len);	// socket is ok
	size_t  	 Read	 (	 char *buf, size_t len);
virtual	void		 Work		(void) = 0;
	bool		 doInternal	(int event);
	void		 Close		(void);
private:void		 printError	(int err);
	void		 doUnblock	(void);
	void		 Init0		(void);
	void		 Init		(void);
	void		 doAccept	(void);
public:	void		 doRead  	(void);
	void		 doWork  	(int event);
virtual bool		 canWork	(void) { return false; }
virtual void		 readWork	(void) { }
virtual	int   	 	 doRead_high		(void) = 0;
	void		 doSend		(void);
	struct in_addr	 My_IP		(void);
	void		 resetAccept	(void);	// close an aktive fd_new
static	void 		 hdlEvent  	(void);
	void		 addEvent	(int msk);
	};

void waitEvent(int usec);

class cSocket *socket_First (void);
class cSocket *socket_Next  (void);
void initSocket(void);
void stopSocket(void);
const char *txtType (const cSocket *list);

#endif

