#include <stdio.h>

#include "misc.h"
#include "protocol.h"
#include "donkey.h"
#include "sGui.h"
#include "sFile.h"
#include "sPacket.h"
#include "sServer.h"
#include "sCommand.h"
#include "sTag.h"
#include "cServer_udp.h"
#include "opcodes.h"

/*
Spezial Hashes:
fffffffffffffffffffffff00000001 -> Information how many dl File Transfers requested
fffffffffffffffffffffff00000002 -> Socket  / Sources
fffffffffffffffffffffff00000003 -> Traffic / Downlaod

Information how many sSource are Slot waiting (max 257 Parts 0 => Have Slot 1-256 => dl_slot+1
*/
static uint16_t sourceCount4Part (tHash &hash, int block, unsigned limit) {{{
	class sFile *file = forHash(hash);
	if (file == NULL) return 0;
	uint16_t ret = file->part_sources[block];
	if (ret > limit) return limit;
	return ret;
	}}}

uint16_t sourceCount4Part (tHash &hash, uint32_t von, uint32_t bis, unsigned limit) {{{
	int block_0 = von / PARTSIZE;
	int block_n = bis / PARTSIZE;
	class sFile *file = forHash(hash);
	if (file == NULL) return 0;
	size_t check = 0;
	for (int j = block_0 ; j <= block_n; j++) check+=file->part_sources[j];
	if (check > limit) check = limit;
	return check;
	}}}

unsigned 	 sGui::count = 0;
sGui *sGui::first_gui = NULL;

sGui     :: sGui      (class cSocket *Fd, struct in_addr ) : sSocket(Fd) {{{
	prio = 0;
	next_gui = first_gui;
	first_gui = this;
	Type = tGui;
	status 	    = CONNECTING;
	name 	    = NULL;
	result_tree = NULL;
	result_len  = 0;
	search_tree = NULL;
	search_len  = 0;
	last_code   = 0;
	auth 	    = false;
	dl_status   = false;
	ul_status   = false;
	announced_server_connect = true ; /* true to send initial disconncet() */
	dl_timeout  = currentTime;
	ul_timeout  = currentTime;
	count++;
	}}}
sGui     ::~sGui      (void) {{{
	if (result_tree != NULL) Free2 (result_tree);
	if (search_tree != NULL) Free2 (search_tree);
	if (name        != NULL) Free2 (name);
	count--;
	// remove me from the file_list
	if (first_gui == this) {
		first_gui = next_gui;
		return;
		}
	sGui *prev_file = first_gui;
	while (prev_file->next_gui != this && prev_file->next_gui != NULL)
	prev_file = prev_file->next_gui;
	if (prev_file == NULL) return;
	ASSERT (prev_file->next_gui == this);
	prev_file->next_gui = prev_file->next_gui->next_gui;
	}}}
void sGui::Name       (const char *NAME) {{{
	Free2 (name);
	name = STRDUP (NAME);
	}}}
void sGui::resultTree	(char *tree,uint32_t len) {{{
	if (result_tree != NULL) Free2 (result_tree);
	result_tree = tree;
	result_len = len;
	}}}
void sGui::searchTree	(char *tree,uint32_t len) {{{
	if (result_tree != NULL) Free2 (search_tree);
	search_tree = tree;
	search_len = len;
	}}}

int   sGui::GUI_txt_g ( void ) {{{
	if (!auth) return 0;
	char logbuf[256];
	GUI_statusMessage("(open)");
	snprintf(logbuf, 255, "ID: %32s", hash_bin2hex(client_hash));
	GUI_statusMessage(logbuf);
	snprintf(logbuf, 255, "%s:%u  (0.0.0.0:0)\n", inet_ntoa(my_ip), pref.ports.client);
	GUI_statusMessage(logbuf);
	snprintf(logbuf, 255, "Users: 172032   Files: 62914");
	GUI_statusMessage(logbuf);
	GUI_statusMessage("Shibboleth: This is cdonkey-0.8.5-cvs hybrid");
	return 1;
	}}}

// Client-Gui    TCP handling (0x64) ... (0xE3)
int   sGui::GUI_serverlist      (void                    ) {{{ // (0xAA) // send server list to gui
	if (!auth) return 0;
	tHash	hash;
	memset (hash, 0, HASH_LEN);
	const struct eServers *s;
	// If all server are send AN if all server use max name.length and max desc.lenth
	unsigned max_cnt = cntServer() ;
	unsigned limit = max_cnt * (78 + sizeof(s->name) + sizeof(s->desc));
	char *BUFFER = (char*)MALLOC (limit);
	char *buffer = BUFFER;
	s = nextServers (true);
	unsigned unused = limit;
	unsigned cnt = 0 ;
	while (s != NULL) {	// we have reserved the maximum possible size
		cnt++;
		ADD_HASH   (&buffer, &unused, hash);
		ADD_U4     (&buffer, &unused, s->ip.s_addr);
		ADD_U2     (&buffer, &unused, s->port     );
		ADD_U4     (&buffer, &unused, 6);
		add_intTag (&buffer, &unused, FT_PRIORITY	, s->prio );
		add_intTag (&buffer, &unused, ST_PING    	, currentTime - s->last_action);
		add_intTag (&buffer, &unused, ST_LONG_USERS	, s->user);
		add_intTag (&buffer, &unused, ST_LONG_FILES	, s->file);
		add_strTag (&buffer, &unused, ST_SERVERNAME	, s->name);
		add_strTag (&buffer, &unused, ST_DESCRIPTION	, s->desc);
		s = nextServers (false);
		}
	log (1, "Server list %i/%i\n", max_cnt, cnt);
    	if (logSend ('>', this, CO_SERVER_LIST, true)) printf("count: %i\n", cnt);
	class sPacket *packet = new sPacket (CO_SERVER_LIST, 4 + limit - unused); // OK
	if (packet == NULL) {{{
		printf("ERROR %s %i\n", __FILE__, __LINE__);
		return 0;
		}}}
	packet->ADD_U4   (cnt);
	if (cnt > 0) packet->ADD_DATA (BUFFER, limit - unused);
	packet->Send(this);
	Free2 (BUFFER);
	delete packet;
	return 1;
	}}}
int   sGui::GUI_shareFiles      (void                    ) {{{ // (0xAD)
    	logSend ('>', this, CO_SHARED_FILES, false);
    	if (!auth) return 0;
    	unsigned limit = 2000; // hardcoded ! BAD
    	char *BUFFER = (char*)MALLOC (limit);
    	char *buffer = BUFFER;
    	unsigned unused = limit;
    	unsigned cnt = 0 ;
    	FILE_LIST {{{
        if ( ! akt->Completed() ) continue;
        	cnt++;
        	ADD_HASH   (&buffer, &unused, akt->hash);   			// hash
        	ADD_U4     (&buffer, &unused, 0);       			// unused
        	ADD_U2     (&buffer, &unused, 0     );  			// unused
        	ADD_U4     (&buffer, &unused, 4);       			// cnt of tag elements
        	add_strTag (&buffer, &unused, FT_FILENAME  , akt->Name());	// filename
        	add_strTag (&buffer, &unused, FT_FILEFORMAT, "unknown");	// fileformat
        	add_intTag (&buffer, &unused, FT_FILESIZE  , akt->Size());	// filesize 
        	add_strTag (&buffer, &unused, FT_FILETYPE  , "unknown" );	// filetype
    		}}}
    	if (cnt > 0) {
        	class sPacket *packet = new sPacket (CO_SHARED_FILES, 4 + limit - unused); // OK
	       	if (packet == NULL) {{{
			printf("ERROR %s %i\n", __FILE__, __LINE__);
			return 0;
			}}}
        	packet->ADD_U4   (cnt);
        	packet->ADD_DATA (BUFFER, limit - unused);
		packet->Send(this);
		delete packet;
    		}
    	Free2 (BUFFER);
    	Code(0xE0); // ? 
    	return 1;
    	}}}
bool  sGui::GUI_res_part_status (tHash  &hash            ) {{{ // (0xAE) SPLIT
	if (!auth) return 0;
	logSend ('>', this, CO_GAP_DETAILS, false);
	sFile *file = forHash (hash);
	if (file == NULL) {{{
		printf("%s %i Unkown file\n", __FILE__, __LINE__);
		return false;
		}}}
	log (1,"stat for %s\n", file->Name());
	unsigned maxCount = (file->size + PARTSIZE - 1) / PARTSIZE;
	uint32_t realCount = 0;
	size_t len , LEN = len = 24 + 10 * maxCount;
	char *buf, *BUF = buf = (char*)MALLOC(LEN);
	char *PartInfo = buf + 24;
	for (unsigned b = 0; b < maxCount; b++) {{{
		uint32_t *p_von   = (uint32_t*)(PartInfo+realCount*10 + 0);
		uint32_t *p_bis   = (uint32_t*)(PartInfo+realCount*10 + 4);
		uint16_t *p_stat  = (uint16_t*)(PartInfo+realCount*10 + 8);
		uint32_t *p1_bis  = (uint32_t*)(PartInfo+realCount*10 + 4 - 10);
		uint16_t *p1_stat = (uint16_t*)(PartInfo+realCount*10 + 8 - 10);
		*p_von = b * PARTSIZE;
		*p_bis = *p_von  + PARTSIZE - 1;
		if (file->partStatus(b)) *p_stat = 2000;
		else {
			if (file->downloads == 0) *p_stat = 0;
			else *p_stat = sourceCount4Part (hash, *p_von, *p_bis, 255);
			}
		if (b == 0) {{{ // continue with next
			realCount++;
			continue;
			}}}
		if (	*p_von - 1 == *p1_bis && *p_stat == *p1_stat) {
			*p1_bis = *p_bis;
			continue;
			}

		realCount++;
		}}}
	log (1,"maxCount / realCount %6u / %6u\n", maxCount, realCount);
	ADD_U1   (&buf, &len, OP_EDONKEYHEADER);
	ADD_U4   (&buf, &len, HASH_LEN + 2 + 10 * realCount + 1);
	ADD_U1   (&buf, &len, CO_GAP_DETAILS);
	ADD_HASH (&buf, &len, hash);
	ADD_U2   (&buf, &len, realCount);
	Write (BUF, 24 + realCount * 10);
	Free2 (BUF);
	return true;
	}}} 
int   sGui::GUI_core_status     (void                    ) {{{ // (0xAF)
	if (!auth) return 0;
	logSend('>', this, CO_CLIENT_STATS, false);
	unsigned long need = 0;
	FILE_LIST need += akt->size / 1024;
	class sPacket *packet = new sPacket (0xAF, 20); // OK
	packet->ADD_F4 (10000.0);				// avail Space on Temp
	packet->ADD_F4 (10000.0); 				// avail Space on Incomming
	packet->ADD_F4 (((float)(need / 1024.0)));		// Space needed by the Downloads
	packet->ADD_U4 (my_ip.s_addr);				// my ID
	packet->ADD_U2 (sSource::count + sServer::count);
	packet->ADD_U2 (10);					// People in the qeue
	packet->Send(this);
	delete packet;
	return 1;
	}}}
int   sGui::GUI_statusMessage   (const char *txt         ) {{{ // (0xB4)
	if (!auth) return 0;
	size_t len = strlen(txt);
	if (len == 0) return 0;
	// ed2k_gui does not need a trailing LF
	char *temp = STRDUP(txt);
	if ( temp[len-1] == '\n' ) {
		temp[len-1] = 0;
		--len;
		}
	if (logSend('>', this, CO_STATUS_MSG, true)) printf("%s\n", txt);
	len = strlen(temp);
	if (len == 0 ) {
		free (temp);
		return 1;
		}
	class sPacket *packet = new sPacket (CO_STATUS_MSG, 2 + len); // OK
	packet->ADD_S2 (temp);
	packet->Send(this);
	free (temp);
	delete packet;
	return 1;
	}}}
int   sGui::GUI_errorMessage    (const char *txt         ) {{{ // (0xB5) Currently not used
	if (!auth) return 0;
	logSend('>', this, CO_ERROR_MSG, false);
	class sPacket *packet = new sPacket (CO_ERROR_MSG, 2 + strlen(txt)); // OK
	packet->ADD_S2 (txt);
	packet->Send(this);
	delete packet;
	return 1;
	}}}
int   sGui::GUI_connected_to    (const char* to_name     ) {{{ // (0xB6)
	if (!auth) return 0;
	logSend('>', this, CO_CONNECTED_TO, false);
	class sPacket *packet = new sPacket (CO_CONNECTED_TO, 2 + strlen(to_name)); // OK
	if (packet == NULL) {{{
		printf("ERROR %s %i\n", __FILE__, __LINE__);
		return 0;
		}}}
	packet->ADD_S2 (to_name);
	packet->Send(this);
	announced_server_connect = true;
	return 1;
	}}}
int   sGui::GUI_disconnected    (void                    ) {{{	// (0xB7)
	if (!auth) return 0;
	/* Dont send disconnect twice */
	if ( announced_server_connect == false ) return 1;
	logSend('>', this, CO_DISCONNECTED, false);
	class sPacket *packet = new sPacket (CO_DISCONNECTED, 0 );
	packet->Send(this); // OK
	delete packet;
	announced_server_connect = false;
	return 1;
	}}}
int   sGui::GUI_result_tcp      (void                    ) {{{ // (0xBB)
	if (!auth) return 0;
	if (resultLen() <= 0) return 0;
	logSend('>', this, CO_ED_NEW_SEARCH_RESULT, false);
	class sPacket *packet = new sPacket (CO_ED_NEW_SEARCH_RESULT, resultLen()); // OK
	packet->ADD_DATA (resultTree(), resultLen());
	packet->Send(this);
	delete packet;
	return 1;
	}}}
int   sGui::GUI_new_download    (const class sFile *file ) {{{ // (0xBC 188)
	if (!auth) return 0;
	logSend('>', this, CO_NEW_DOWNLOAD, false);
	const char *tmp_name = file->Name();
	const char *tmp_hash = (const char*)hash_bin2hex (file->hash);
	size_t len = 43 + strlen(tmp_name) + strlen(tmp_hash);
	class sPacket *packet = new sPacket (CO_NEW_DOWNLOAD, len); // OK
	packet->ADD_HASH (file->hash);
	packet->ADD_U4 ( 0 );			// IP
	packet->ADD_U2 ( 0 );			// PORT
	packet->ADD_U4 ( 2 );			// number of tags
	add_strTag (packet, FT_FILENAME, tmp_name);
	add_intTag (packet, FT_FILESIZE, file->Size());
	packet->ADD_U1 ( 1 ); 			// Prio
	packet->ADD_S2 ( tmp_hash ); 		// uniqID
	packet->Send(this);
	delete packet;
	return 1;
	}}}
int   sGui::GUI_new_upload      (const class sFile *file ) {{{ // (0xBE 190)
	if (!auth) return 0;
	logSend('>', this, CO_NEW_UPLOAD, false);
	const char *tmp_name = "Group";
	size_t len = 2 + strlen(file->Name()) + 40 + strlen(tmp_name);
	class sPacket *packet = new sPacket (CO_NEW_UPLOAD, len); // OK
	packet->ADD_S2   (file->Name());
	packet->ADD_HASH (file->hash);
	packet->ADD_U4   ( 0 );			// IP
	packet->ADD_U2   ( 0 );			// Port
	packet->ADD_U4   ( 2 );			// number of tags
	add_strTag (packet, FT_FILENAME, tmp_name);
	add_intTag (packet, FT_FILESIZE, file->Size());
	packet->Send(this);
	delete packet;
	return 1;
	}}}
int   sGui::GUI_removeDownload  (tHash &hash             ) {{{ // (0xBD)
	if (!auth) return 0;
	if (!validSocket()) return 0;
	if (logSend('>', this, CO_REMOVE_DOWNLOAD, true)) printf("%32s\n", hash_bin2hex(hash));
	class sPacket *packet = new sPacket (CO_REMOVE_DOWNLOAD, HASH_LEN); //OK
	packet->ADD_HASH(hash);
	packet->Send(this);
	delete packet;
	return 1;
	}}}
int   sGui::GUI_dl_status       (void                    ) {{{ // (0xC5 197)
	if (!auth) return 0;
	logSend('>', this, CO_DOWNLOAD_STATUS, false);
	int cnt = sFile::count - sFile::count_complete;
	int len = 2 + 13 * cnt;
	class sPacket *packet = new sPacket (CO_DOWNLOAD_STATUS, len); //OK
	if (packet == NULL) {{{
		printf("ERROR %s %i\n", __FILE__, __LINE__);
		return 0;
		}}}
	packet->ADD_U2( cnt) ; 			// NUM of DL
	uint16_t dl_id = 0;
	FILE_LIST {{{				// Information about all files
		if ( akt->Completed() ) continue; // dont show finished files in main view
		packet->ADD_U2 (dl_id++); 		// DL_ID
		uint8_t akt_status = STATUS_LOOKING;
		uint8_t akt_sources = 255;
		if (akt->downloads < akt_sources) akt_sources = akt->downloads;
		if (akt->downloads 	> 0   	  ) akt_status = STATUS_DOWNLOADING;
		if (akt->Pause  () 	      	  ) akt_status = STATUS_PAUSED;
		if (akt->status == ALL_DONE	  ) akt_status = STATUS_DONE;
		packet->ADD_U1 ( akt_status);				// DL_STATUS
		packet->ADD_F4 (((float)(akt->dlRate()/1024.0))) ; 	// current speed
		packet->ADD_U4 ( akt->transfered ) ; 		// kb transferred
		unsigned not_full = 0;
		unsigned trying   = 0;
		unsigned maxCount = (akt->size + PARTSIZE - 1) / PARTSIZE;
		tHash hash;
		memcpy (hash, akt->hash , HASH_LEN);
		for (unsigned i = 0; i < maxCount; i++) {
			if (akt->partStatus(i)) continue;
			not_full ++;
			if (sourceCount4Part (hash, i, 1)) trying++;
			}
		uint8_t percent = 100;
		if (not_full > 0) percent = (trying * 100) / not_full;
//		if (akt->Size() > 0) percent = akt->transfered * 100 / akt->Size();
		packet->ADD_U1 (percent); 				// % avalable
		packet->ADD_U1 (akt_sources);		// sources (max 255 shown) count OK but Gui show other value WHY ?
		}}}
	packet->Send(this);
	delete packet;
	dl_timeout = currentTime;
	return 1;
	}}}
int   sGui::GUI_ul_status       (void                    ) {{{ // (0xC6 218) Mehr Infos would be nice
	if (!auth) return 0;
	logSend('>', this, CO_UPLOAD_STATUS, false);
	int cnt = sFile::count;
	if (cnt > 65535) cnt = 65535;
	int len = 2 + 6 * cnt;
	class sPacket *packet = new sPacket (CO_UPLOAD_STATUS, len); // OK
	if (packet == NULL) {{{
		printf("ERROR %s %i\n", __FILE__, __LINE__);
		return 0;
		}}}
	packet->ADD_U2 (((uint16_t)cnt)); 			// NUM of DL
	uint16_t ul_id = 0;
	FILE_LIST {						// LOOP through all files
		float rate = akt->ulRate();
		rate /= 1024.0;
		packet->ADD_U2 (ul_id++); 	// UL_ID-1 since here is no extra info
		packet->ADD_F4 (rate);		// current speed
		}
	packet->Send(this);
	delete packet;
	ul_timeout = currentTime;
	return 1;
	}}}
int   sGui::GUI_send_option     (void                    ) {{{ // (0xC7)
	if (!auth) return 0;
	logSend ('>', this, CO_OPTIONS, false);
	size_t len = 43 + strlen(client_name) + strlen(incoming) + strlen(completed) ;
	class sPacket *packet = new sPacket (CO_OPTIONS, len); // OK
	if (packet == NULL) {{{
		printf("ERROR %s %i\n", __FILE__, __LINE__);
		return 0;
		}}}
	packet->ADD_U2 (((uint16_t)my_version)); 		// client version
	packet->ADD_F4 (((float)(dlSecond / 1024.0)));	// max down
	packet->ADD_F4 (((float)(ulSecond / 1024.0)));	// max up
	packet->ADD_U2 (pref.ports.client);			// doorport
	packet->ADD_U2 (0);				// CRAP (deprecated)
	packet->ADD_S2 (client_name);			// client name
	packet->ADD_S2 (incoming);			// temp dir
	packet->ADD_S2 (completed);			// incoming dir (completed dir)
	packet->ADD_U1 (0);				// DUMMY if overnet
	packet->ADD_U1 (0);				// DUMMY if overnet
	packet->ADD_U1 (1);				// allow private message
	packet->ADD_U1 (1);				// save corrupt
	packet->ADD_U1 (1);				// DUMMY 
	packet->ADD_U2 (pref.ports.gui);		// admin port
	packet->ADD_U4 (TCP_CLIENT_COUNT) ;		// max connections
	packet->ADD_U4 (0x020c0800);			// core build date
	packet->ADD_F4(((float)(dlSecond / 1024.0)));	// line download speed
	packet->ADD_U4 (getpid ());			// PID
	packet->Send(this);
	delete packet;
	Code(0x01);
	return 1;
	}}}
int   sGui::GUI_share_dirs      (void                    ) {{{ // (0xD7) // sende meine shared dirs an die GUI
	logSend('>', this, CO_SHARED_DIRS, false);
	if (!auth) return 0;
	if (completed == NULL ) return 0;
	size_t len = 2 + 2 + strlen(completed);
	class sPacket *packet = new sPacket (CO_SHARED_DIRS, len); // OK
	if (packet == NULL) {{{
		printf("ERROR %s %i\n", __FILE__, __LINE__);
		return 0;
		}}}
	packet->ADD_U2( 1 );
	packet->ADD_S2 (completed);
	packet->Send(this);
	delete packet;
	return 1;
	}}}

int  sGui::GUI_login            (sPacket *packet) {{{ // (0x64)
	logSend ('<', this, CO_LOGIN, false);
	if (packet == NULL) return 0;
	if (ctrl_pass == NULL || ctrl_user == NULL) return 0;
	char *log_name = packet->GET_S2 ();
	char *pass = packet->GET_S2();
	if ( 0 != strcmp (log_name, ctrl_user) || 0 != strcmp (pass, ctrl_pass)) {{{
		printf("Gui DENY:%s pass:%s\n", log_name, pass );
		Free2 (log_name);
		Free2 (pass);
		Close();
		return 0;
		}}}
	auth = true;
	Code(0x64);
	Free2 (log_name);
	Free2 (pass);
	FILE_LIST {
		if ( ! akt->Completed() ) GUI_new_download (akt);
		GUI_new_upload(akt);
		}
	if ( tcp_server != NULL ) {
		// Stati which will be ignored for successful connection
		if ( tcp_server->Status() >= 4 && tcp_server->Status() <= 15 
			&& tcp_server->Status() != SERVER_CLOSED 
			&& tcp_server->Status() != WAIT_SERVER_LIST ) {
			struct eServers *s = Server_ip (tcp_server->peer_ip);
			if ( s != NULL ) {
				char tmpname[33];
				printf("XXX connected to %15s,%s Status: %i \n", inet_ntoa( s->ip ), s->name, tcp_server->Status() );
			  	if ( strlen( s->name ) == 0 ) {
					// s->name ist leider nicht immer gesetzt (BUG?). Wir verwenden dann die IP
					// This is legal if the server have not yet respond with its name
					snprintf(tmpname,33, "%15s", inet_ntoa( s->ip ) );	
					GUI_connected_to(tmpname);
					}
			  	else GUI_connected_to(s->name);
				}
			}
		}
	return 1;
	}}}
int  sGui::GUI_shutdown         (sPacket *      ) {{{ // (0x65)
	logSend ('<', this, CO_SHUTDOWN, false);
	if (!auth) return 0;
	running = false;
	return 1;
	}}}
int  sGui::GUI_txt_command      (sPacket *packet) {{{ // (0x66)
	logSend ('<', this, CO_TXTCOMMAND, false);
	if (packet == NULL) return 0;
	if (!auth) return 0;
	char *text = packet->GET_S2 ();
	const char *msg = sCommand::Execute(text);
	GUI_statusMessage (msg);
	Free2 (text);
	return 1;
	}}}
int  sGui::GUI_connect          (sPacket *packet) {{{ // (0xC8)
	if (packet == NULL) return 0;
	struct in_addr  ip;
	uint16_t	new_port;
	char logbuf[256];
	ip.s_addr = packet->GET_U4();
	new_port	  = packet->GET_U2();
	if (ip.s_addr == 0) {
		if (logSend('<', this, CO_CONNECT, true)) printf("RANDOM\n");
		struct in_addr  new_ip;
		if (tcp_server != NULL) delete tcp_server;
		GUI_disconnected();
		if (nextServer (&new_ip, &new_port)) {
			tcp_server = new sServer (new_ip, new_port);
			snprintf(logbuf,255, "Trying Edonkey Server: %15s\n", inet_ntoa(new_ip) );
            		log (1,logbuf);
            		GUI_LIST akt->GUI_statusMessage(logbuf);
			}
		return 1;
		}
	if (logSend('<', this, CO_CONNECT, true)) printf("SELECTED %15s:%5u\n", inet_ntoa(ip), new_port);
	if (tcp_server != NULL) delete tcp_server;
	GUI_disconnected();
	tcp_server = new sServer (ip, new_port);
	snprintf(logbuf,255, "Trying Edonkey Server: %15s\n", inet_ntoa(ip) );
	log (1,logbuf);
	GUI_LIST akt->GUI_statusMessage(logbuf); 
	return 1;
	}}}
int  sGui::GUI_disconnect       (sPacket *      ) {{{ // (0xC9)
	logSend ('<', this, CO_DISCONNECT, false);
	if (tcp_server != NULL) delete tcp_server ; //tcp_server->Close();
	tcp_server = NULL ;
	GUI_disconnected();
	return 1;
	}}}
int  sGui::GUI_search           (sPacket *packet) {{{ // (0xCA)
	if (packet == NULL) return 0;
	logSend ('<', this, CO_SEARCH, false);
	uint32_t len = packet->Length() - 1;
	char *tree = (char*)MALLOC (len);
	memcpy (tree, packet->DATA(), len);
	packet->GET_U1 ();
	char *searchstr = packet->GET_S2 ();
	log (1,"Searching: %s\n", searchstr) ;
	free(searchstr);
	searchTree (tree, len);
	tcp_search = true;
	return 1;
	}}}
int  sGui::GUI_search_ext       (sPacket *      ) {{{ // (0xCB)
	logSend ('<', this, CO_EXSEARCH, false);
	userver_listen->UDP_search4gui(this);
	return 1;
	}}}
int  sGui::req_download         (sPacket *packet) {{{ // (0xCF)
	logSend ('<', this, CO_DOWNLOAD, false);
	tHash hash;
	packet->GET_HASH(hash);
	if (NULL != forHash(hash)) return 1;	//dl already exists
	printf("HASH %32s\n", hash_bin2hex(hash));
	const struct eFiles *f = nextFile(hash, true);
	printf("Found %i\n", cursor_count);
	while (f != NULL) {
		if (f->info_len != 0) {
			size_t len =  f->info_len;
			char *dat = (char*)MALLOC(len);
			memcpy(dat, f->info, len);
			sTag *tag = parse_tag (dat, &len);
			if (tag == NULL) {
				Free2 (dat);
				printf("bad\n");
				break;
				}
			Free2 (dat);
			if (tag->size > 0 && tag->name != NULL) {
				sFile *file = new sFile (hash, tag->name, tag->size, incoming);
	        		GUI_new_download (file);
					GUI_new_upload (file);
				delete tag;
				return 1;
				}
			delete tag;
			}
		f = nextFile(hash, false);
		}
	return 1;
	}}}
int  sGui::GUI_dlPause          (sPacket *packet) {{{ // (0xD0)
	if (packet == NULL) return 0;
	logSend ('<', this, CO_PAUSE_DOWNLOAD, false);
	tHash hash;
	packet->GET_HASH (hash);
	sFile *file = forHash (hash);
	if (file == NULL) return 1;
	file->Pause (1);
	return 1;
	}}}
int  sGui::GUI_dlResume         (sPacket *packet) {{{ // (0xD1)
	if (packet == NULL) return 0;
	logSend ('<', this, CO_RESUME_DOWNLOAD, false);
	tHash hash;
	packet->GET_HASH (hash);
	sFile *file = forHash (hash);
	if (file == NULL) return 1;
	file->Pause (0);
	return 1;
	}}}
int  sGui::GUI_cancelDownload   (sPacket *packet) {{{ // (0xD2)
	if (packet == NULL) return 0;
	logSend ('<', this, CO_CANCEL_DOWNLOAD, false);
	tHash hash;
	packet->GET_HASH (hash);
	sFile *file = forHash (hash);
	delete file;
	char *alt = (char*)MALLOC (strlen(control) + 38);
	char *neu = (char*)MALLOC (strlen(control) + 42);
	sprintf (alt, "%s%c%s.lst"    , control, dir_delim, hash_bin2hex (hash));
	sprintf (neu, "%s%c%s.lst.del", control, dir_delim, hash_bin2hex (hash));
	rename(alt, neu);
	Free2 (neu);
	Free2 (alt);
	return 1;
	}}}
int  sGui::GUI_download_prio    (sPacket *packet) {{{ // (0xD3)
	if (packet == NULL) return 0;
	logSend ('<', this, CO_SETPRI_DOWNLOAD, false);
	tHash hash;
	packet->GET_HASH (hash);
	uint8_t sprio = packet->GET_U1 ();
	printf("Prio %u\n",sprio);
	if (sprio>2) return 1;
	sFile *file = forHash (hash);
	if (file == NULL) return 1;
	file->dlPrio (sprio);
	return 1;
	}}}
int  sGui::GUI_get_serverlist   (sPacket *      ) {{{ // (0xD5) // Gui request for server List
	if (!auth) return 0;
	logSend ('<', this, CO_GET_SERVERLIST, false);
	GUI_serverlist ();
	return 1;
	}}}
int  sGui::GUI_req_shared_dir   (sPacket *      ) {{{ // (0xD7)
	if(!auth) return 0;
	logSend ('<', this, CO_GET_SHARE_DIRS, false);
	GUI_share_dirs ();
	return 1;
	}}}
int  sGui::GUI_dlStatus_start   (sPacket *      ) {{{ // (0xD9)
	logSend ('<', this, CO_START_DL_STATUS, false);
	dl_status  = true;
	if(!auth) return 0;
	GUI_dl_status();
	return 1;
	}}}
int  sGui::GUI_dlStatus_stop    (sPacket *      ) {{{ // (0xDA)
	logSend ('<', this, CO_STOP_DL_STATUS, false);
	if(!auth) return 0;
	dl_status  = false;
	return 1;
	}}}
int  sGui::GUI_ulStatus_start   (sPacket *      ) {{{ // (0xDB)
	if(!auth) return 0;
	logSend ('<', this, CO_START_UL_STATUS, false);
	ul_status  = true;
	GUI_ul_status();
	return 1;
	}}}
int  sGui::GUI_ulStatus_stop    (sPacket *      ) {{{ // (0xDC)
	if(!auth) return 0;
	logSend ('<', this, CO_STOP_UL_STATUS, false);
	ul_status  = false;
	return 1;
	}}}
int  sGui::GUI_serverDelete     (sPacket *packet) {{{ // (0xDD)
	if (packet == NULL || !auth) return 0;
	logSend ('<', this, CO_DELETE_SERVER, false);
	struct in_addr  ip;
	uint16_t	new_port;
	ip.s_addr = packet->GET_U4 ();
	new_port  = packet->GET_U2 ();
	delServer (ip, new_port, -7);
	return 1;
	}}}
int  sGui::GUI_serverAdd        (sPacket *packet) {{{ // (0xDE)
	if (packet == NULL || !auth) return 0;
	logSend ('<', this, CO_ADD_SERVER, false);
	struct in_addr  ip;
	uint16_t	new_port;
	ip.s_addr = packet->GET_U4 ();
	new_port  = packet->GET_U2 ();
	saveServer (ip, new_port, 0, 0, true);
	return 1;
	}}}
int  sGui::GUI_serverPrio       (sPacket *packet) {{{ // (0xDF)
	if (packet == NULL) return 0;
	logSend ('<', this, CO_SETPRI_SERVER, false);
	if(!auth) return 0;
	struct in_addr  ip;
	ip.s_addr = packet->GET_U4 ();
	uint16_t new_port = packet->GET_U2 ();
	uint8_t  new_prio = packet->GET_U1 ();
	setServer_prio (ip, new_port, new_prio);
	return 1;
	}}}
int  sGui::GUI_get_shareFiles   (sPacket *      ) {{{ // (0xE0)
	if (!auth) return 0;
	logSend ('<', this, CO_GET_SHARE_FILES, false);
	GUI_shareFiles();
	Code(0xE0); // ? 
	return 1;
	}}}
int  sGui::GUI_get_option       (sPacket *      ) {{{ // (0xE1)
	if (!auth) return 0;
	logSend ('<', this, CO_GET_OPTIONS, false);
	GUI_send_option ();
	return 1;
	}}}
int  sGui::GUI_add_download     (sPacket *packet) {{{ // (0xE2)
	if (packet == NULL) return 0;
	logSend ('<', this, CO_REQ_NEW_DOWNLOAD, false);
	tHash hash;
	packet->GET_HASH(hash);
	if (NULL != forHash(hash)) return 1; // dl already exists
	packet->GET_U4  (    );
	packet->GET_U2  (    );
	sTag *tag = parseTag (packet);
	if (tag == NULL) return 0;
	if (tag->name != NULL && tag->size > 0)
		new sFile (hash, tag->name, tag->size, incoming);
	delete tag;
	return 1;
	}}}
int  sGui::GUI_req_part_status  (sPacket *packet) {{{ // (0xE3)
	if (packet == NULL) return 0;
	logSend ('<', this, CO_GET_GAP_DETAILS, false);
	tHash hash;
	packet->GET_HASH(hash);
	return GUI_res_part_status (hash);
	}}}
int  sGui::GUI_core_status_get	 (sPacket *      ) {{{ // (0xe4)
	if (!auth) return 0;
	return GUI_core_status ();
	}}}

int sGui::Donkey_gui (class sPacket *packet) {{{
	if (packet == NULL) return 0;
	if (packet->Proto() != OP_EDONKEYHEADER) return 0;
	switch (packet->Command ()) {
//	case 0xCC : More Search
//	case 0xCD : Search User
//	case 0xD8 : Set Share Dirs
//	case 0xCE : ext Search user
	case 0xD6 : return 1; // Get Friend List (if anyone wan't it)
//	case 0xD4 : View Friend Files
//	case 0xE4 : return 1; // Request Client Stats
	case 0x64 : return GUI_login		(packet);
	case 0x65 : return GUI_shutdown		(packet);
	case 0x66 : return GUI_txt_command	(packet);
	case 0xC8 : return GUI_connect   	(packet);
	case 0xC9 : return GUI_disconnect	(packet);
	case 0xCA : return GUI_search		(packet);
	case 0xCB : return GUI_search_ext	(packet);
	case 0xCF : return req_download 	(packet);
	case 0xD0 : return GUI_dlPause		(packet); // Pause  Download <HASH>
	case 0xD1 : return GUI_dlResume		(packet); // Resume Download <HASH>
	case 0xD2 : return GUI_cancelDownload	(packet);
	case 0xD3 : return GUI_download_prio	(packet); // Prio   Download <HASH><8 bit>
	case 0xD5 : return GUI_get_serverlist	(packet);
	case 0xD7 : return GUI_req_shared_dir	(packet);
	case 0xD9 : return GUI_dlStatus_start	(packet);
	case 0xDA : return GUI_dlStatus_stop	(packet);
	case 0xDB : return GUI_ulStatus_start	(packet);
	case 0xDC : return GUI_ulStatus_stop 	(packet);
	case 0xDD : return GUI_serverDelete	(packet);
	case 0xDE : return GUI_serverAdd	(packet);
	case 0xDF : return GUI_serverPrio	(packet);
	case 0xE0 : return GUI_get_shareFiles	(packet);
	case 0xE1 : return GUI_get_option	(packet);
	case 0xE2 : return GUI_add_download	(packet);
	case 0xE3 : return GUI_req_part_status	(packet);
	case 0xE4 : return GUI_core_status_get	(packet);
	default   : logSend ('<', this, packet->Command (), false);
		    break;
		}
	packet->Dump ();
	return 0;
	}}}

void sGui::readWork(void) {{{
	if (READ->Proto() == OP_EDONKEYHEADER) Donkey_gui (READ);
	else 	READ->Dump();
	READ->readReset();
	}}}

void    sGui::Work (void) {{{
	if (!validSocket()) return;
	if(!auth) return;
	if (dl_timeout + 3 < currentTime || dl_timeout > currentTime) dl_timeout = currentTime;
	if (ul_timeout + 3 < currentTime || ul_timeout > currentTime)  ul_timeout = currentTime;
	if(dl_timeout + 1 < currentTime && dl_status) GUI_dl_status();
	if(ul_timeout + 1 < currentTime && ul_status) GUI_ul_status();
	return;
	}}}
/*
 * vim600: fdm=marker
 */
